/*
    Tcl.js
    Proof of concept put together by Patrick Gillespie (http://patorjk.com/)
    BSD Licensed
    
    This interpreter is a port of Picol by Salvatore antirez Sanfilippo.
    From the Picol source code file: 
    
    Tcl in ~ 500 lines of code by Salvatore antirez Sanfilippo. BSD licensed
    URL: http://antirez.com/page/picol.html
*/

"use strict";

var Tcl = Tcl || (function() {

    var PICOL_OK =       {toString: function(){return "PICOL_OK";}},// toString was done for debugging reasons
        PICOL_ERR =      {toString: function(){return "PICOL_ERR";}},
        PICOL_RETURN =   {toString: function(){return "PICOL_RETURN";}},
        PICOL_BREAK =    {toString: function(){return "PICOL_BREAK";}},
        PICOL_CONTINUE = {toString: function(){return "PICOL_CONTINUE";}},
        PT_ESC = {toString: function(){return "PT_ESC";}},
        PT_STR = {toString: function(){return "PT_STR";}},
        PT_CMD = {toString: function(){return "PT_CMD";}},
        PT_VAR = {toString: function(){return "PT_VAR";}},
        PT_SEP = {toString: function(){return "PT_SEP";}},
        PT_EOL = {toString: function(){return "PT_EOL";}},
        PT_EOF = {toString: function(){return "PT_EOF";}};

    var looseNodes = {};

    /*
        Begin utility functions
    */

    function picolInitParser(parser, text) {
        //p = {};
        parser.text = text;
        parser.p = 0;
        parser.len = text.length;
        parser.start = 0;
        parser.end = 0;
        parser.insidequote = 0;
        parser.type = PT_EOL;
    }

    function picolParseSep(parser) {
        parser.start = parser.p;
        while(  parser.text[parser.p] === " " || parser.text[parser.p] === "\t" || 
                parser.text[parser.p] === "\n" || parser.text[parser.p] === "\r") {
            parser.p++;
            parser.len--;
        }
        parser.end = parser.p-1;
        parser.type = PT_SEP;
        return PICOL_OK;
    }

    function picolParseEol(parser) {
        parser.start = parser.p;
        while ( parser.text[parser.p] === " " || parser.text[parser.p] === "\t" || 
                parser.text[parser.p] === "\n" || parser.text[parser.p] === "\r" ||
                parser.text[parser.p] === ";")
        {
            parser.p++;
            parser.len--;
        }
        parser.end = parser.p - 1;
        parser.type = PT_EOL;
        return PICOL_OK;
    }

    function picolParseCommand(parser) {
        var level = 1,
            blevel = 0;
        parser.start = ++parser.p;
        parser.len--;
        while (1) {
            if (parser.len === 0) {
                break;
            } else if (parser.text[parser.p] === "[" && blevel === 0) {
                level++;
            } else if (parser.text[parser.p] === "]" && blevel === 0) {
                if (!--level) break;
            } else if (parser.text[parser.p] === "\\") {
                parser.p++;
                parser.len--;
            } else if (parser.text[parser.p] === "{") {
                blevel++;
            } else if (parser.text[parser.p] === "}") {
                if (blevel !== 0) blevel--;
            }
            parser.p++;
            parser.len--;
        }
        parser.end = parser.p - 1;
        parser.type = PT_CMD;
        if (parser.text[parser.p] === "]") {
            parser.p++;
            parser.len--;
        }
        return PICOL_OK;
    }

    function picolParseVar(parser) {
        parser.start = ++parser.p;
        parser.len--;
        while (1) {
            if ((parser.text[parser.p] >= "a" && parser.text[parser.p] <= "z") || 
                (parser.text[parser.p] >= "A" && parser.text[parser.p] <= "Z") ||
                (parser.text[parser.p] >= "0" && parser.text[parser.p]<= "9") || (parser.text[parser.p] === "_")) 
            {
                parser.p++;
                parser.len--;
                continue;
            }
            break;
        }
        if (parser.start === parser.p) {
            parser.start = parser.end = parser.p - 1;
            parser.type = PT_STR;
        } else {
            parser.end = parser.p - 1;
            parser.type = PT_VAR;
        }
        return PICOL_OK;
    }

    function picolParseBrace(parser) {
        var level = 1;
        parser.start = ++parser.p;
        parser.len--;
        while (1) {
            if (parser.len >= 2 && parser.text[parser.p] === "\\") {
                parser.p++;
                parser.len--;
            } else if (parser.len === 0 || parser.text[parser.p] === "}") {
                level--;
                if (level === 0 || parser.len === 0) {
                    parser.end = parser.p-1;
                    if (parser.len) {
                        parser.p++;
                        parser.len--;
                    }
                    parser.type = PT_STR;
                    return PICOL_OK;
                }
            } else if (parser.text[parser.p] === "{") {
                level++;
            }
            parser.p++;
            parser.len--;
        }
        return PICOL_OK; // unreached
    }

    function picolParseString(parser) {
        var newword = (parser.type === PT_SEP || parser.type === PT_EOL || parser.type === PT_STR);
        if (newword && parser.text[parser.p] === "{") return picolParseBrace(parser);
        else if (newword && parser.text[parser.p] == '"') {
            parser.insidequote = 1;
            parser.p++; 
            parser.len--;
        }
        parser.start = parser.p;
        while(1) {
            if (parser.len === 0) {
                parser.end = parser.p - 1;
                parser.type = PT_ESC;
                return PICOL_OK;
            }
            switch(parser.text[parser.p]) {
            case '\\':
                if (parser.len >= 2) {
                    parser.p++; 
                    parser.len--;
                }
                break;
            case '$': 
            case '[':
                parser.end = parser.p - 1;
                parser.type = PT_ESC;
                return PICOL_OK;
            case ' ': 
            case '\t': 
            case '\n': 
            case '\r': 
            case ';':
                if (!parser.insidequote) {
                    parser.end = parser.p - 1;
                    parser.type = PT_ESC;
                    return PICOL_OK;
                }
                break;
            case '"':
                if (parser.insidequote) {
                    parser.end = parser.p - 1;
                    parser.type = PT_ESC;
                    parser.p++; 
                    parser.len--;
                    parser.insidequote = 0;
                    return PICOL_OK;
                }
                break;
            }
            parser.p++; 
            parser.len--;
        }
        return PICOL_OK; /* unreached */
    }

    function picolParseComment(parser) {
        while(parser.len && parser.text[parser.p] != '\n') {
            parser.p++; 
            parser.len--;
        }
        return PICOL_OK;
    }

    function picolGetToken(parser) {
        debug("Entering picolGetToken");
        while(1) {
            if (!parser.len) {
                if (parser.type !== PT_EOL && parser.type !== PT_EOF)
                    parser.type = PT_EOL;
                else
                    parser.type = PT_EOF;
                return PICOL_OK;
            }
            switch(parser.text[parser.p]) {
            case ' ': 
            case '\t': 
            case '\r':
                if (parser.insidequote) return picolParseString(parser);
                return picolParseSep(parser);
            case '\n': 
            case ';':
                if (parser.insidequote) return picolParseString(parser);
                return picolParseEol(parser);
            case '[':
                return picolParseCommand(parser);
            case '$':
                return picolParseVar(parser);
            case '#':
                if (parser.type == PT_EOL) {
                    picolParseComment(parser);
                    continue;
                }
                return picolParseString(parser);
            default:
                return picolParseString(parser);
            }
        }
        return PICOL_OK; // unreached
    }

    function picolCreateInterp() {
        var interpreter = {};
        interpreter.level = 0;
        interpreter.callframe = {};
        interpreter.callframe.vars = null;
        interpreter.callframe.parent = null;
        interpreter.commands = null;
        interpreter.result = "";
        return interpreter;
    }

    function picolSetResult(interpreter, s) {
        interpreter.result = s;
    }

    function picolGetVar(interpreter, name) {
        var v = interpreter.callframe.vars;
        while(v) {
            if (v.name === name) return v;
            v = v.next;
        }
        return null;
    }

    function picolSetVar(interpreter, name, val) {
        var v = picolGetVar(interpreter, name);
        if (v) {
            v.val = val;
        } else {
            v = {};
            v.name = name;
            v.val = val;
            v.next = interpreter.callframe.vars;
            interpreter.callframe.vars = v;
        }
        return PICOL_OK;
    }

    function picolGetCommand(interpreter, name) {
        var c = interpreter.commands,
            elm;
        while(c) {
            if (c.name === name) return c;
            c = c.next;
        }
        
        // no registered command found, are we dealing with a dom node?
        if ( elm = getDomElement(name) ) {
            c = {};
            c.func = picolCommandDomElement;
            return c;
        }
        
        return null;
    }

    function picolRegisterCommand(interpreter, name, cmdFunct, privdata) {
        var cmd = picolGetCommand(interpreter, name);
        if (cmd) {
            console.log("Command " + name + " is already defined.");
            picolSetResult(interpreter,"Command " + name + " is already defined.");
            return PICOL_ERR;
        }
        cmd = {};
        cmd.name = name;
        cmd.func = cmdFunct;
        cmd.privdata = privdata;
        cmd.next = interpreter.commands;
        interpreter.commands = cmd;
        return PICOL_OK;
    }

    /* EVAL! */

    function picolEval(interpreter, text) {
        var parser = {},
            argc = 0,
            j,
            argv = {},
            errbuf,
            retcode = PICOL_OK,
            v,
            token,
            tlen,
            prevtype,
            cmd;
            
        picolSetResult(interpreter, "");
        picolInitParser(parser, text);
        while(1) {
            prevtype = parser.type
            picolGetToken(parser);
            if (parser.type === PT_EOF) break;
            tlen = parser.end - parser.start + 1;
            if (tlen < 0) tlen = 0;
            token = {};
            token = parser.text.substr(parser.start, tlen);
            debug("token="+token);
            debug("parser.type="+parser.type);
            if (parser.type == PT_VAR) {
                v = picolGetVar(interpreter, token);
                if (!v) {
                    errbuf = "No such variable " + token;
                    console.log(errbuf);
                    token = "";
                    picolSetResult(interpreter, errbuf);
                    retcode = PICOL_ERR;
                    break;
                }
                token = v.val;
            } else if (parser.type == PT_CMD) {
                retcode = picolEval(interpreter, token);
                token = "";
                if (retcode != PICOL_OK) break;
                token = interpreter.result;
            } else if (parser.type == PT_ESC) {
                // XXX: escape handling missing!
            } else if (parser.type == PT_SEP) {
                prevtype = parser.type;
                token = "";
                continue;
            }
            // We have a complete command + args. Call it!
            if (parser.type == PT_EOL) {
                token = "";
                prevtype = parser.type;
                if (argc) {
                    if ((cmd = picolGetCommand(interpreter,argv[0])) === null) {
                        errbuf = "No such command " + argv[0];
                        console.log(errbuf);
                        picolSetResult(interpreter, errbuf);
                        retcode = PICOL_ERR;
                        break;
                    }
                    retcode = cmd.func(interpreter, argc, argv, cmd.privdata);
                    if (retcode != PICOL_OK) break;
                }
                // Prepare for the next command
                argv = {};
                argc = 0;
                continue;
            }
            // We have a new token, append to the previous or as new arg?
            if (prevtype === PT_SEP || prevtype === PT_EOL) {
                debug("argc="+argc);
                debug("token="+token);
                argv[argc] = token;
                argc++;
            } else { // Interpolation
                var oldlen = argv[argc-1].length, 
                    tlen = token.length;
                argv[argc-1] += token;
                token = "";
            }
            prevtype = parser.type;
        }
    err:
        argv = null;
        return retcode;
    }

    /* ACTUAL COMMANDS! */
    
    function picolArityErr(interpreter, name) {
        var buf;
        buf = "Wrong number of args for " + name;
        console.log(buf);
        picolSetResult(interpreter,buf);
        return PICOL_ERR;
    }

    function picolCommandMath(interpreter, argc, argv, pd) {
        var buf, a, b, c;
        if (argc !== 3) return picolArityErr(interpreter,argv[0]);
        a = Number(argv[1]); 
        b = Number(argv[2]);
        if (argv[0][0] === '+') c = a+b;
        else if (argv[0][0] === '-') c = a-b;
        else if (argv[0][0] === '*') c = a*b;
        else if (argv[0][0] === '/') c = a/b;
        else if (argv[0][0] === '>' && argv[0][1] === '\0') c = a > b;
        else if (argv[0][0] === '>' && argv[0][1] === '=') c = a >= b;
        else if (argv[0][0] === '<' && argv[0][1] === '\0') c = a < b;
        else if (argv[0][0] === '<' && argv[0][1] === '=') c = a <= b;
        else if (argv[0][0] === '=' && argv[0][1] === '=') c = a === b;
        else if (argv[0][0] === '!' && argv[0][1] === '=') c = a !== b;
        else c = 0; // I hate warnings
        buf = c;
        picolSetResult(interpreter,buf);
        return PICOL_OK;
    }

    function picolCommandSet(interpreter, argc, argv, pd) {
        if (argc !== 3) return picolArityErr(interpreter,argv[0]);
        picolSetVar(interpreter,argv[1], argv[2]);
        picolSetResult(interpreter, argv[2]);
        return PICOL_OK;
    }

    function picolCommandPuts(interpreter, argc, argv, pd) {
        if (argc != 2) return picolArityErr(interpreter,argv[0]);
        console.log(argv[1]);
        return PICOL_OK;
    }

    function picolCommandIf(interpreter, argc, argv, pd) {
        var retcode;
        if (argc !== 3 && argc !== 5) return picolArityErr(interpreter, argv[0]);
        if ((retcode = picolEval(interpreter,argv[1])) !== PICOL_OK) return retcode;
        if (Number(interpreter.result)) return picolEval(interpreter, argv[2]);
        else if (argc === 5) return picolEval(interpreter,argv[4]);
        return PICOL_OK;
    }

    function picolCommandWhile(interpreter, argc, argv, pd) {
        if (argc !== 3) return picolArityErr(interpreter, argv[0]);
        while(1) {
            var retcode = picolEval(interpreter,argv[1]);
            if (retcode !== PICOL_OK) return retcode;
            if (Number(interpreter.result)) {
                if ((retcode = picolEval(interpreter,argv[2])) === PICOL_CONTINUE) continue;
                else if (retcode === PICOL_OK) continue;
                else if (retcode === PICOL_BREAK) return PICOL_OK;
                else return retcode;
            } else {
                return PICOL_OK;
            }
        }
    }

    function picolCommandRetCodes(interpreter, argc, argv, pd) {
        if (argc !== 1) return picolArityErr(interpreter,argv[0]);
        if (argv[0] === "break") return PICOL_BREAK;
        else if (argv[0] === "continue") return PICOL_CONTINUE;
        return PICOL_OK;
    }

    function picolDropCallFrame(interpreter) {
        var cf = interpreter.callframe;
        var v = cf.vars;
        var t;
        while(v) {
            t = v.next;
            v = t;
        }
        interpreter.callframe = cf.parent;
        cf = null;
    }

    /*
        interpreter
        argc - argument count
        argv - argument varaibles
        pd - procedure data, [0] - arguments list, [1] - procedure body
    */
    function picolCommandCallProc(interpreter, argc, argv, pd) {
        var argListStr = pd[0], 
            body = pd[1], 
            pos = 0, 
            cf = {},
            arity = 0, 
            done = 0, 
            retCode = PICOL_OK,
            errbuf,
            errNow = false;
            
        cf.vars = null;
        cf.parent = interpreter.callframe;
        interpreter.callframe = cf;
        
        while(1) {
            var start = pos;
            while(argListStr[pos] !== ' ' && argListStr.length > pos) pos++;
            if (pos < argListStr.length && pos === start) {
                pos++; continue;
            }
            if (pos === start) break;
            if (pos === argListStr.length) done = 1;
            if (++arity > argc-1) {
               errNow = true;
               break;
            }
            picolSetVar(interpreter, argListStr.substring(start,pos), argv[arity]);
            pos++;
            if (done) break;
        }
        
        if (arity === argc-1) {
            retCode = picolEval(interpreter,body);
            if (retCode == PICOL_RETURN) retCode = PICOL_OK;
            picolDropCallFrame(interpreter); // remove the called proc callframe
            return retCode;
        } else {
            errbuf = "Proc '" + argv[0] + "' called with wrong arg num";
            picolSetResult(interpreter, errbuf);
            picolDropCallFrame(interpreter); // remove the called proc callframe
            return PICOL_ERR;
        }
    }

    function picolCommandProc(interpreter, argc, argv, pd) {
        var procdata = ["", ""];
        if (argc !== 4) return picolArityErr(interpreter,argv[0]);
        procdata[0] = argv[2]; // arguments list 
        procdata[1] = argv[3]; // procedure body
        return picolRegisterCommand(interpreter, argv[1], picolCommandCallProc, procdata);
    }

    function picolCommandComment() {
        //do nothing
        return PICOL_OK;
    }

    function getDomElement(str) {
        if ( /^domNode_/.test(str) ) {
            if (looseNodes[str]) {
                return looseNodes[str];
            } else {
                return document.getElementById(str.replace(/^domNode_/, ""));
            }
        }
        return null;
    }

    function picolCommandWindow(interpreter, argc, argv, pd) {
        var elm,
            domId,
            tclId;
        if (argv[1] === "tclSupport" && argc === 2) {
            picolSetResult(interpreter, window.tclSupport);
        } else if (argv[1] === "tclSupport" && argc === 3) {
            window.tclSupport = argv[2];
        }        
        return PICOL_OK;
    }

    function picolCommandDocument(interpreter, argc, argv, pd) {
        var elm,
            domId,
            tclId;
        if (argc === 3) {
            if (argv[1] === "getElementById") {
                picolSetResult(interpreter, "domNode_" + argv[2]);
            } else if (argv[1] === "createTextNode") {
                elm = document.createTextNode(argv[2]);
                elm.id = "node" + Math.floor(Math.random()*1000000000);
                tclId = "domNode_" + elm.id;
                looseNodes[tclId] = elm;
                picolSetResult(interpreter, tclId);
            } else if (argv[1] === "appendChild") {
                if ( elm = getDomElement(argv[2]) ) {
                    document.appendChild(elm);
                    if (looseNodes[argv[2]]) {
                        delete looseNodes[argv[2]];
                    }
                }
            }
        }
        return PICOL_OK;
    }

    function picolCommandDomElement(interpreter, argc, argv, pd) {
        var targetElm = getDomElement(argv[0]),
            elm,
            domId,
            tclId;
        if (targetElm && argc === 3) {
            if (argv[1] === "appendChild") {
                if ( elm = getDomElement(argv[2]) ) {
                    targetElm.appendChild(elm);
                    if (looseNodes[argv[2]]) {
                        delete looseNodes[argv[2]];
                    }
                }
            }
        }
        return PICOL_OK;
    }

    function picolCommandReturn(interpreter, argc, argv, pd) {
        if (argc != 1 && argc != 2) return picolArityErr(interpreter,argv[0]);
        picolSetResult(interpreter, (argc == 2) ? argv[1] : "");
        return PICOL_RETURN;
    }

    function picolRegisterCoreCommands(interpreter) {
        var j, name = ["+","-","*","/",">",">=","<","<=","==","!="];
        for (j = 0; j < name.length; j++)
            picolRegisterCommand(interpreter, name[j], picolCommandMath, null);
        picolRegisterCommand(interpreter, "set", picolCommandSet, null);
        picolRegisterCommand(interpreter, "puts", picolCommandPuts, null);
        picolRegisterCommand(interpreter, "if", picolCommandIf, null);
        picolRegisterCommand(interpreter, "while", picolCommandWhile, null);
        picolRegisterCommand(interpreter, "break", picolCommandRetCodes, null);
        picolRegisterCommand(interpreter, "continue", picolCommandRetCodes, null);
        picolRegisterCommand(interpreter, "proc", picolCommandProc, null);
        picolRegisterCommand(interpreter, "return", picolCommandReturn, null);
        picolRegisterCommand(interpreter, "#", picolCommandComment, null);
        picolRegisterCommand(interpreter, "document", picolCommandDocument, null);
        picolRegisterCommand(interpreter, "window", picolCommandWindow, null);
    }

    var isDebugging = false;
    function debug(txt) {
        if (isDebugging) {
            console.log(txt);   
        }
    }

    // what we return
    var myTcl = {};
    myTcl.eval = function(code) {
        var interp = picolCreateInterp();
        picolRegisterCoreCommands(interp);

        if (picolEval(interp, code) != PICOL_OK) {
            console.log("An error occurred :/");
            console.log(interp.result)
        };
    };
    
    return myTcl;
})();